/* NEC Arcadia GbitEthernet driver
 *
 * (C) Copyright 2006 Hiroyasu Nakatsuka
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 *    REVISION HISTORY:
 *    ================
 *
 *    v0.1       1-6-2006      nakatuka        Suport eth0 only
 *
 */

#include <config.h>


#ifdef CONFIG_ARCADIA

#include <common.h>
#include <malloc.h>
#include <net.h>
#include <command.h>
#include <asm/io.h>
#include <asm/cacheops.h>
#include <asm/ywingreg.h>

/* Ethernet PHY address */
#define ET0_PHY_ADDR 0x00
#define ET1_PHY_ADDR 0x01

/* descriptor count */
#define NUM_TX_DESC 4
#define NUM_RX_DESC PKTBUFSRX

/* descriptor pointer */
static int next_rx;
static int now_rx;

#define NEXTTXDESC(n) ((n)<NUM_TX_DESC-1?(n)+1:0)
#define NEXTRXDESC(n) ((n)<NUM_RX_DESC-1?(n)+1:0)

typedef struct{
	u32 addr;
	u16 size;
	u16 flag;
} DESC;

static DESC rx_descriptor[NUM_RX_DESC+1];
static DESC tx_descriptor[NUM_TX_DESC+1];

struct arcadia_hw {
	void *hw_addr;
	void *mdio_addr;
	uint32_t phy_addr;
	int media;
};

#define MAX_WAIT 1000

/* media type selection */
#define MEDIA_NONE   	0
#define MEDIA_10_HALF   1
#define MEDIA_10_FULL   2
#define MEDIA_100_HALF  3
#define MEDIA_100_FULL  4
#define MEDIA_1000_HALF 5
#define MEDIA_1000_FULL 6

static int rx_desc_in(struct arcadia_hw *hw);
static int ether_adr_init(struct eth_device* dev);
static void ether_mii_reset(struct arcadia_hw *hw);
static void ether_hw_reset(struct arcadia_hw *hw);
static int ether_int_init(struct arcadia_hw *hw);
static int ether_mac_init(struct arcadia_hw *hw);
static void ethdsc_init(struct arcadia_hw *hw);
static void mdio_write(struct arcadia_hw* hw, int phyreg_no, int data);
static unsigned int mdio_read(struct arcadia_hw* hw, int phyreg_no);
static void ether_link_status(struct arcadia_hw* hw, int reg, int reg2);
static void ether_auto_negotiation(struct arcadia_hw* hw);
extern volatile uchar	PktBuf[(PKTBUFSRX+1) * PKTSIZE_ALIGN + PKTALIGN];

/* L1 D-cache writeback macro */
static inline void L1_D_writeback(ulong addr)
{
  __asm__ __volatile__ (
    ".set noreorder\n\t"
    "cache %1,  (%0)\n\t"
    "cache %1, 1(%0)\n\t"
	"cache %1, 2(%0)\n\t"
	"cache %1, 3(%0)\n\t"
    " .set reorder"
    :
    : "r" (addr),
      "i" (Hit_Writeback_D)
  );
}


/**************************************************************************/
/* SEND - Transmit a frame                                                */
/***************************************************************************/
static int arcadia_send(struct eth_device* dev, volatile void *packet, int length){
	struct arcadia_hw* hw = (struct arcadia_hw*)dev->priv;
	volatile DESC *tx_desc =
                (volatile DESC*)(KSEG1ADDR(&tx_descriptor[0]));
	volatile u32 *gmacc1 = (volatile u32*)(hw->hw_addr+GMACC1);
	volatile u32 *gtx_dpr = (volatile u32*)(hw->hw_addr+GTX_DPR);
	volatile u32 *gtx_cfgr = (volatile u32*)(hw->hw_addr+GTX_CFGR);
	volatile u32 *gtx_mbtsz = (volatile u32*)(hw->hw_addr+GTX_MBTSZ);
	int i, line;
	int res;
	u32 buff_addr;

	/* tx MAC/DMA disable & tx DMA burst size setting */
	*gtx_mbtsz = GDMABST_32;
	*gmacc1 &= ~GTEN;
	*gtx_cfgr &= ~GTXE;
	y_sync();
	udelay(1);

	/* tx data buffer must be writeback */
	buff_addr = (u32)packet;
	line = (length+0x20 -1) / 0x20;
	for(i=0; i<=line; i++){
		 L1_D_writeback(buff_addr);
		 buff_addr += 0x20;
	}
	
    /* tx descriptor should always be idle */
	tx_desc[0].size = length;
	tx_desc[0].flag =  LAST | DB_LP;
	tx_desc[0].addr = (virt_to_phys(packet));
	tx_desc[1].size = 0;      /* NULL descriptor */
	tx_desc[1].flag = 0;
	tx_desc[1].addr = 0;

	y_sync();
	udelay(1);
//	printf("desc[0] %x %x %x \n", tx_desc[0].size, tx_desc[0].flag, tx_desc[0].addr);
	
	/* tx DMA enable */
	*gtx_dpr = (virt_to_phys(&tx_desc[0]));
	*gmacc1 |= GTEN;
	*gtx_cfgr = GTXE;
	y_sync();
//	printf("send2 gmacc1:%x gtx_cfgr:%x \n", *gmacc1, *gtx_cfgr);
	
	/* wait Tx Done */
	udelay(1);
	i=0;
    while(!(tx_desc[0].flag & OWN)){
		if(i>MAX_WAIT){
			printf("TX timeout:%x \n", (tx_desc[0].flag));
			break;
		}
		udelay(10);
		i++;
	}
	udelay(10);

    /* Clear done bit */
	tx_desc[0].size = 0;
    tx_desc[0].addr = 0;

	y_sync();

	res = tx_desc[0].flag;
	return(res);
}

/**************************************************************************/
/* RECV - Receive a frame                                                 */
/**************************************************************************/
/***** data output from now rx-descriptor to data buffer *****/
/* return value
 *  0: receive is OK
 *  1: nothing has been received
 *  2: receive data is ERROR
 */
static int arcadia_recv(struct eth_device* dev)
{
	struct arcadia_hw* hw = (struct arcadia_hw*)dev->priv;
	volatile DESC *rx_desc =
                (volatile DESC*)(KSEG1ADDR(&rx_descriptor[0]));
	volatile u32 *gmacc1 = (volatile u32*)(hw->hw_addr+GMACC1);
	volatile u32 *grx_cfgr = (volatile u32*)(hw->hw_addr+GRX_CFGR);
	volatile u32 *grx_mbtsz = (volatile u32*)(hw->hw_addr+GRX_MBTSZ);
	volatile u32 *grx_enmon = (volatile u32*)(hw->hw_addr+GRX_ENMON);
	int idx;
	int length;
	u16 status;

	/* if rx DMA is not enable, enable rx DMA & rx DMA burst size setting */
	if(*grx_enmon & GRXESTANDBY){
		*grx_mbtsz = GDMABST_32;
		*gmacc1 |= GREN;
		*grx_cfgr = GRXE;
	}

	for(;;){
		idx = now_rx;
   /* available rx desctiptor does not exist */
		if(idx == next_rx) return 1;

   /* nothing has been received */
		if(!(rx_desc[idx].flag & OWN)) return 1;

   /* receive OK */
		status = rx_desc[idx].flag;
		length = rx_desc[idx].size;
		if( !(status & G_RXOK)) {
			printf("[Rx Error] status:%04x \n", status);
			return 2;
		}
		/* Pass the packet up to the protocol layyers */
		else {
			/* printf("[%d] status:%04x \n", idx, status); */
			NetReceive((void*) KSEG1ADDR(NetRxPackets[idx]), length - 4);
		}

   /* free now descriptor */
		rx_desc[idx].flag = 0;
		rx_desc[idx].size = 0;
		rx_desc[idx].addr  = 0;
		now_rx = NEXTRXDESC(idx);
		if( rx_desc_in(hw)) return 2;
		return 0;
	}
}

/**************************************************************************/
/* INIT - Look for an adapter, this routine's visible to the outside      */
/**************************************************************************/
static int arcadia_init(struct eth_device* dev, bd_t * bd)
{
	struct arcadia_hw* hw = (struct arcadia_hw*)dev->priv;
	int i;

	ether_adr_init(dev);
	ether_hw_reset(hw);

	ether_mii_reset(hw);
	udelay(100);

	if ( hw->media == MEDIA_NONE )
	  ether_auto_negotiation(hw);

	ether_mac_init(hw);
	ether_int_init(hw);
	ethdsc_init(hw);

	/* make rx descriptor */
	for(i=0;i<CFG_RX_ETH_BUFFER;i++){
		rx_desc_in(hw);
	}

	return 1;
}

/*====================  arcadia MAC Unit initialize  ===================*/
static int ether_mac_init(struct arcadia_hw* hw)
{
	volatile u32 *gmacc1 = (volatile u32*)(hw->hw_addr+GMACC1);
	volatile u32 *gmacc2 = (volatile u32*)(hw->hw_addr+GMACC2);
	volatile u32 *gmacc3 = (volatile u32*)(hw->hw_addr+GMACC3);
	volatile u32 *gmiscr = (volatile u32*)(hw->hw_addr+GMISCR);
	volatile u32 *ght1 = (volatile u32*)(hw->hw_addr+GHT1);
	volatile u32 *ght2 = (volatile u32*)(hw->hw_addr+GHT2);
	volatile u32 *gafr = (volatile u32*)(hw->hw_addr+GAFR);
	volatile u32 *grfic2 = (volatile u32*)(hw->hw_addr+GRFIC2);
	volatile u32 *grfic3 = (volatile u32*)(hw->hw_addr+GRFIC3);
	volatile u32 *gtfic1 = (volatile u32*)(hw->hw_addr+GTFIC1);

	if(hw->media ==  MEDIA_1000_FULL){
  // GMII, No loopback, No flow control(except Tx for BugFix), Full duplex
		*gmacc1 =  GTFCEN;
		*gmacc2 =  GPRELN | GIFMOD | GPADEN | GCRCEN | GFULLD;
	} else if(hw->media ==  MEDIA_1000_HALF){
  // GMII, No loopback, No flow control(except Tx for BugFix), Half duplex
		*gmacc1 =  GTFCEN;
		*gmacc2 =  GPRELN | GIFMOD | GPADEN | GCRCEN;
	} else if((hw->media ==  MEDIA_100_FULL)||(hw->media ==  MEDIA_10_FULL)){
  // No loopback, No flow control(except Tx for BugFix), Full duplex
		*gmacc1 =  GTFCEN;
		*gmacc2 =  GPRELN | GIFMOD_4 | GPADEN | GCRCEN | GFULLD;
	} else {
  // No loopback, No flow control(except Tx for BugFix), Half duplex
		*gmacc1 =  GTFCEN;
		*gmacc2 =  GPRELN | GIFMOD_4 | GPADEN | GCRCEN;
	}

	*gmiscr = GPSEL | GPEN; // Odd parity enable
	*gmacc3 = GAPSE;        // Status Info append to tail of packet 
	*ght1 = 0x00000000;     // no hash table use
	*ght2 = 0x00000000;
	*gafr = GPRO;           // All packet recieve mode( Promiscuous mode )
	*grfic2 = 0x00000000;   // All packet receieve (disable packet discard)
	*grfic3 = GTHRX | GRXETH; // THRX:no use(default), RXETH:no use(default)
	*gtfic1 = GTFWMH | GTFDWL;// TFWMH:0x1200(default), TFDWL:0x0600(default)

	y_resetw((GSRST|GRMCRST|GTMCRST|GRFRST|GTFRST), 10, GTFCEN, gmacc1); /* MACC1 soft reset */
	
	return 0;
}

/*=====================  arcadia Int initialize  ====================*/
static int ether_int_init(struct arcadia_hw *hw){
	volatile u32 *gtimr = (volatile u32*)(hw->hw_addr+GTIMR);
	volatile u32 *grimr = (volatile u32*)(hw->hw_addr+GRIMR);
	volatile u32 *gtsvreg = (volatile u32*)(hw->hw_addr+GTSVREG);
	volatile u32 *grsvreg = (volatile u32*)(hw->hw_addr+GRSVREG);
	volatile u32 *gfsvreg = (volatile u32*)(hw->hw_addr+GFSVREG);
	volatile u32 *gtx_msr = (volatile u32*)(hw->hw_addr+GTX_MSR);
	volatile u32 *grx_msr = (volatile u32*)(hw->hw_addr+GRX_MSR);
	volatile u32 *gtx_isr = (volatile u32*)(hw->hw_addr+GTX_ISR);
	volatile u32 *grx_isr = (volatile u32*)(hw->hw_addr+GRX_ISR);

	*gtimr = 0x00000000; // All Tx int masked (MAC)
	*grimr = 0x00000000; // All Rx int masked (MAC)
	*gtsvreg = 0x00000000; // All Tx int status clear (MAC)
	*grsvreg = 0x00000000; // All Rx int status clear (MAC)
	*gfsvreg = 0x00000000; // All FIFO int status clear (MAC)
	*gtx_msr = 0x00000000; // All Tx int masked (DMA)
	*grx_msr = 0x00000000; // All Rx int masked (DMA)
	*gtx_isr = GTXDRBUSERR | GTXDWBUSERR | GTXDTBUSERR | GTXDNDESC | GTXDESCNULL | GTXEDEA | GTXHL; // All Tx int status clear (DMA)
	*grx_isr = GRXDRBUSERR | GRXDWBUSERR | GRXDTBUSERR | GRXDNDESC | GRXDAL | GRXDZR | GRXEDEA | GRXLFAL | GRXHL; // All Rx int status clear (DMA)
	
	return 0;
}

/*====================  set MAC address  =============================*/
static int ether_adr_init(struct eth_device* dev){
	struct arcadia_hw* hw = (struct arcadia_hw*)dev->priv;
	volatile u32 *glsa1 = (volatile u32*)(hw->hw_addr+GLSA1);
	volatile u32 *glsa2 = (volatile u32*)(hw->hw_addr+GLSA2);
	u32 mac_addr_hi, mac_addr_lo;

        mac_addr_hi  =	(dev->enetaddr[1] << 24) | (dev->enetaddr[0] << 16) ;
        mac_addr_lo  =	(dev->enetaddr[5] << 24) | (dev->enetaddr[4] << 16) |
			(dev->enetaddr[3] <<  8) | (dev->enetaddr[2]      ) ;
	/*
	printf( "ether_adr_init: %x:%x:%x:%x:%x:%x\n",
			(mac_addr_hi >> 16) & 0xFF,
			(mac_addr_hi >> 24) & 0xFF,
			(mac_addr_lo )      & 0xFF,
			(mac_addr_lo >> 8)  & 0xFF,
			(mac_addr_lo >> 16) & 0xFF,
			(mac_addr_lo >> 24) & 0xFF);
	*/
	*glsa2 = mac_addr_hi;
	*glsa1 = mac_addr_lo;
		
	return 0;
}

/*==============  Tx/Rx DMA unit initialization  =====================*/
static void ether_hw_reset(struct arcadia_hw *hw){
	volatile u32 *gtx_srst = (volatile u32*)(hw->hw_addr+GTX_SRST);
	volatile u32 *grx_srst = (volatile u32*)(hw->hw_addr+GRX_SRST);
	
	y_resetw(GTXRST, 10, 0x00000000, gtx_srst);/* Tx-DMA soft reset */
	y_resetw((GMACRST|GRXRST), 10, 0x00000000, grx_srst);/* Rx-DMA soft reset */
}

/*================  MII unit initialization  =======================*/
static void ether_mii_reset(struct arcadia_hw* hw){
	volatile u32 *gmacc3 = (volatile u32*)(hw->hw_addr+GMACC3);
	volatile u32 *gmacc1 = (volatile u32*)(hw->hw_addr+GMACC1);
	volatile u32 *gstlc = (volatile u32*)(hw->hw_addr+GSTLC);
	volatile u32 *gmiic = (volatile u32*)(hw->hw_addr+GMIIC);
	volatile u32 *miic = (volatile u32*)(hw->mdio_addr+MIIC);
	
	y_resetw((GTXFFLH|GRXFFLH), 10, 0x00000000, gmacc3);/* MACC3 soft reset */
	y_resetw((GSRST|GRMCRST|GTMCRST|GRFRST|GTFRST), 10, GTFCEN, gmacc1); /* MACC1 soft reset */
	y_resetw(GSTRST, 10, 0x00000000, gstlc);  /* STLC soft reset */
	y_resetw(GMISRT, 10, GCLKS108, gmiic);  /* MII block soft reset */
	y_resetw((MISRT|CLKS66), 10, CLKS66, miic);  /* MII block soft reset(ch2) */
}

/**************************************************************************/
/* MDIO treatment                                                         */
/**************************************************************************/
/***** auto negotiation *****/
static void ether_auto_negotiation(struct arcadia_hw* hw)
{
	int reg,reg2,bAutoNego,i;

	mdio_write(hw, MII_CONTROL, MIICNTL_AUTO);  /* ANEGEN */
	mdio_write(hw, MII_CONTROL, MIICNTL_AUTO | MIICNTL_RST_AUTO);  /* restart auto nego */
	bAutoNego = 0;

	printf( "Auto Negotiation " );
	for(i=0; i < 5000; i++) {
		reg = mdio_read(hw, MII_STATUS);
		if(reg & MIISTAT_AUTO_DONE) {
			bAutoNego = 1;
			break;
		}
		udelay(1000);
	}

	if(!bAutoNego){                         /* Setup 10BASE-T Half duplex */
		mdio_write(hw, MII_CONTROL, 0x0000);
		printf("failed, 10Base-T Half duplex\n");
	} else {
	  reg = mdio_read(hw, MII_ANLPAR);
	  reg2 = mdio_read(hw, MII_GBSR);
	  ether_link_status(hw, reg, reg2);
	}
}

static void ether_link_status(struct arcadia_hw* hw, int reg, int reg2)
{
	if(reg2 & MII_NWAY_1000_FDX){
		printf("complete: 1000Base Full duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_SPEED1 | MIICNTL_FDX);
		hw->media = MEDIA_1000_FULL;
	}
	else if(reg2 & MII_NWAY_1000){
		printf("complete: 1000Base Half duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_SPEED1);
		hw->media = MEDIA_1000_HALF;
	}		
	else if(reg & MII_NWAY_TX_FDX){
		printf("complete: 100Base Full duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_SPEED | MIICNTL_FDX);
		hw->media = MEDIA_100_FULL;
	}
	else if(reg & MII_NWAY_TX){
		printf("complete: 100Base Half duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_SPEED);
		hw->media = MEDIA_100_HALF;
	}
	else if(reg &  MII_NWAY_T_FDX){
		printf("complete: 10Base Full duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_FDX);
		hw->media = MEDIA_10_FULL;
	}
	else {
	  printf("complete: 10Base Half duplex\n");
	  //mdio_write(hw, MII_CONTROL, 0x0000);
	  hw->media = MEDIA_10_HALF;
	}
}

static unsigned int mdio_read(struct arcadia_hw* hw, int phyreg_no){
	volatile u32 *madr = (volatile u32*)(hw->mdio_addr+MADR);
	volatile u32 *mcmd = (volatile u32*)(hw->mdio_addr+MCMD);
	volatile u32 *mind = (volatile u32*)(hw->mdio_addr+MIND);
	volatile u32 *mrdd = (volatile u32*)(hw->mdio_addr+MRDD);

	/* phy addr & cmd set */
	*madr = (hw->phy_addr <<  FIAD_SHIFT) | phyreg_no;
	*mcmd = RSTAT;

	/* wait not busy = read is comlited */
	while( *mind & BUSY);

	return *mrdd;
}

static void mdio_write(struct arcadia_hw* hw, int phyreg_no, int data){
	volatile u32 *madr = (volatile u32*)(hw->mdio_addr+MADR);
	volatile u32 *mind = (volatile u32*)(hw->mdio_addr+MIND);
	volatile u32 *mwtd = (volatile u32*)(hw->mdio_addr+MWTD);

	/* phy addr & cmd set */
	*madr = (hw->phy_addr <<  FIAD_SHIFT) | phyreg_no;

	/* data write */
	*mwtd = data;

	/* wait not busy = write is comlited */
	while( *mind & BUSY);
}

/**************************************************************************/
/* Descriptor treatment                                                   */
/**************************************************************************/
static void ethdsc_init(struct arcadia_hw* hw)
{
	volatile u32 *grx_dpr = (volatile u32*)(hw->hw_addr+GRX_DPR);
	volatile u32 *gtx_dpr = (volatile u32*)(hw->hw_addr+GTX_DPR);
	volatile DESC *rx_desc =
                (volatile DESC*)(KSEG1ADDR(&rx_descriptor[0]));
	volatile DESC *tx_desc =
				(volatile DESC*)(KSEG1ADDR(&tx_descriptor[0]));
	int i;

//	printf("tx_desc:%x rx_desc:%x \n", &tx_desc[0], &rx_desc[0]);

	for(i=0; i<NUM_RX_DESC; i++){
		rx_desc[i].size = 0;
		rx_desc[i].flag = 0;
		rx_desc[i].addr = 0;
	}
	rx_desc[i].size = 0;
	rx_desc[i].flag = 0;
	rx_desc[i].addr = (virt_to_phys(&rx_desc[0]));

	for(i=0; i<NUM_TX_DESC; i++){
		tx_desc[i].size = 0;
		tx_desc[i].flag = 0;
		tx_desc[i].addr = 0;
	}
	tx_desc[i].size = 0;
	tx_desc[i].flag = 0;
	tx_desc[i].addr = (virt_to_phys(&tx_desc[0])); 

	next_rx = now_rx = 0;
	
	*grx_dpr = (virt_to_phys(&rx_desc[0]));
	*gtx_dpr = (virt_to_phys(&tx_desc[0]));
}


/*******************  make next rx-descriptor  ***********************/
/* return value
 *  0: success make next rx descriptor
 *  1: rx descriptor ring is full
 */
static int rx_desc_in(struct arcadia_hw* hw)
{
	volatile DESC *rx_desc =
                (volatile DESC*)(KSEG1ADDR(&rx_descriptor[0]));
	volatile u32 *grx_pdr = (volatile u32*)(hw->hw_addr+GRX_PDR);
	volatile u32 *grx_plalm = (volatile u32*)(hw->hw_addr+GRX_PLALM);
	int ri,next;

	/* make next rx-pointer set */
	ri = next_rx;
	next = NEXTRXDESC(ri);
	if(next == now_rx) {
		printf("rx descriptor is full \n");
		return 1;
	}

	rx_desc[ri].size = PKTSIZE_ALIGN;
	rx_desc[ri].flag = DB_LP;
	rx_desc[ri].addr = (virt_to_phys(NetRxPackets[ri]));

	if ( NetRxPackets[ri] < &PktBuf[0] ||
	     NetRxPackets[ri] >= &PktBuf[sizeof(PktBuf)-PKTSIZE_ALIGN] )
	  printf( "rx_desc_in(%d): out of buffer: size: %d, adr=%x (%x< %x <%x)\n",
		  ri, rx_desc[ri].size, NetRxPackets[ri], rx_desc[ri].addr,
		  &PktBuf[0], &PktBuf[sizeof(PktBuf)/PKTSIZE_ALIGN] );

	/* remaing descriptor number(RNOD) is plused 1 */
	*grx_pdr = 0x1;
	*grx_plalm = 0x0; /* alerm level is set to 0 */
	
	next_rx = next;
	return 0;
}


/**************************************************************************/
/* HALT - Turn off ethernet interface                                     */
/**************************************************************************/
static void arcadia_halt(struct eth_device* dev){
	struct arcadia_hw* hw = (struct arcadia_hw*)dev->priv;
	volatile u32 *gmacc1 = (volatile u32*)(hw->hw_addr+GMACC1);
	volatile u32 *grx_cfgr = (volatile u32*)(hw->hw_addr+GRX_CFGR);
    /* MAC/DMA rx disable */
	*gmacc1 &= ~GREN;
	*grx_cfgr &= ~GRXE;
}

static int register_arcadia( void *hw_addr, void *mdio_addr, uint32_t phy_addr, unsigned char*mac,
			     const char *post )
{
	struct eth_device* dev;
	struct arcadia_hw* hw;
	int i;

	dev = (struct eth_device*) malloc(sizeof(*dev));
	memset(dev, 0, sizeof(*dev));
	hw = (struct arcadia_hw*) malloc(sizeof(*hw));
	memset(hw, 0, sizeof(*hw));

	sprintf(dev->name, "NEC-Arcadia%s", post);
	dev->iobase = 0;
	dev->priv   = hw;
	dev->init   = arcadia_init;
	dev->halt   = arcadia_halt;
	dev->send   = arcadia_send;
	dev->recv   = arcadia_recv;
	for (i = 0; i < 6; ++i) {
		dev->enetaddr[i] = ((unsigned char*)0xBFFFF100)[i];
	}

	hw->hw_addr = hw_addr;
	hw->mdio_addr = mdio_addr;
	hw->phy_addr = phy_addr;
	eth_register(dev);

	return 1;
}
int arcadia_enet_initialize(bd_t *bis){
	volatile u16 *gpio_sel = (volatile u16*)(GPIO_SEL);
	volatile u16 *gpio_iosel4 = (volatile u16*)(GPIO_IOSEL4);
	volatile u16 *gpio_data4 =  (volatile u16*)(GPIO_DATA4);

	/* set GMII term mode */
	*gpio_sel &= 0xf0ff;
	*gpio_sel |= 0x0c00;
	
	/* GPHY reset disable */
	*gpio_iosel4 |= 0x3c00;
	*gpio_data4 |= 0x2000;

	return register_arcadia( (void*)GETH0_BASE, (void*)ETH1_BASE, ET0_PHY_ADDR,
				 (unsigned char*)0xBFFFF100, "" );
}

#endif /* CONFIG_ARCADIA */
