/* NEC candy ethernet driver
 *
 * (C) Copyright 2005 Hiroyasu Nakatsuka
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 *    REVISION HISTORY:
 *    ================
 *
 *    v0.5       7-7-2005      nakatuka        Suport eth0 only
 *
 */

#include <config.h>


#ifdef CONFIG_CANDY

#include <common.h>
#include <malloc.h>
#include <net.h>
#include <command.h>
#include <asm/io.h>
#include <asm/cacheops.h>
#include <asm/ywingreg.h>

/* Ethernet PHY address */
#define ET0_PHY_ADDR 0x01

/* descriptor count */
#define NUM_TX_DESC 4
#define NUM_RX_DESC PKTBUFSRX

/* descriptor pointer */
static int next_rx;
static int now_rx;

#define NEXTTXDESC(n) ((n)<NUM_TX_DESC-1?(n)+1:0)
#define NEXTRXDESC(n) ((n)<NUM_RX_DESC-1?(n)+1:0)

typedef struct{
	u16 size;
	u16 flag;
	u32 addr;
} DESC;

static DESC rx_descriptor[NUM_RX_DESC+1];
static DESC tx_descriptor[NUM_TX_DESC+1];

struct candy_hw {
	void *hw_addr;
	uint32_t phy_addr;
	int media;
};

#define MAX_WAIT 1000

/* media type selection */
#define MEDIA_NONE   	0
#define MEDIA_10_HALF   1
#define MEDIA_10_FULL   2
#define MEDIA_100_HALF  3
#define MEDIA_100_FULL  4
#define MEDIA_1000_HALF 5
#define MEDIA_1000_FULL 6

static int rx_desc_in(struct candy_hw* hw);
static int ether_adr_init(struct eth_device* dev);
static void ether_mii_reset(struct candy_hw* hw);
static int ether_mac_init(struct candy_hw* hw);
static void ethdsc_init(struct candy_hw* hw);
static void mdio_write(struct candy_hw* hw, int phyreg_no, int data);
static unsigned int mdio_read(struct candy_hw* hw, int phyreg_no);
static void ether_link_status(struct candy_hw* hw, int reg);
static void ether_auto_negotiation(struct candy_hw* hw);
extern volatile uchar	PktBuf[(PKTBUFSRX+1) * PKTSIZE_ALIGN + PKTALIGN];

/* L1 D-cache writeback macro */
static inline void L1_D_writeback(ulong addr)
{
  __asm__ __volatile__ (
    ".set noreorder\n\t"
    "cache %1,  (%0)\n\t"
    "cache %1, 1(%0)\n\t"
	"cache %1, 2(%0)\n\t"
	"cache %1, 3(%0)\n\t"
    " .set reorder"
    :
    : "r" (addr),
      "i" (Hit_Writeback_D)
  );
}


/**************************************************************************/
/* SEND - Transmit a frame                                                */
/***************************************************************************/
static int candy_send(struct eth_device* dev, volatile void *packet, int length){
	struct candy_hw* hw = (struct candy_hw*)dev->priv;
	volatile DESC *tx_desc =
                (volatile DESC*)(KSEG1ADDR(&tx_descriptor[0]));
//	volatile u32 *macc1 = (volatile u32*)(hw->hw_addr+MACC1);
	volatile u32 *xmt_dpr = (volatile u32*)(hw->hw_addr+XMT_DPR);
	volatile u32 *xmt_cfgr = (volatile u32*)(hw->hw_addr+XMT_CFGR);
	int i, line;
	int res;
	u32 buff_addr;

	/* tx DMA disable & tx DMA burst size setting*/
	*xmt_cfgr = DTBS8;
	y_sync();
	udelay(1);

	/* tx data buffer must be writeback */
	buff_addr = (u32)packet;
	line = (length+0x20 -1) / 0x20;
	for(i=0; i<=line; i++){
		 L1_D_writeback(buff_addr);
		 buff_addr += 0x20;
	}
	
    /* tx descriptor should always be idle */
	tx_desc[0].size = length;
	tx_desc[0].flag =  LAST | DB_LP;
    tx_desc[0].addr = (virt_to_phys(packet));
	tx_desc[1].size = 0;      /* NULL descriptor */
	tx_desc[1].flag = 0;
    tx_desc[1].addr = 0;

	y_sync();
	udelay(1);
//	printf("desc[0] %x %x %x \n", tx_desc[0].size, tx_desc[0].flag, tx_desc[0].addr);
	
	/* tx DMA enable */
	*xmt_cfgr = TXE | DTBS8;
	*xmt_dpr = (virt_to_phys(&tx_desc[0]));
	y_sync();
	
	/* wait Tx Done */
	udelay(1);
	i=0;
    while(!(tx_desc[0].flag & OWN)){
		if(i>MAX_WAIT){
			printf("TX timeout:%x \n", (tx_desc[0].flag));
			break;
		}
		udelay(10);
		i++;
	}
	udelay(10);

    /* Clear done bit */
	tx_desc[0].size = 0;
    tx_desc[0].addr = 0;

	y_sync();

	res = tx_desc[0].flag;
	return(res);
}

/**************************************************************************/
/* RECV - Receive a frame                                                 */
/**************************************************************************/
/***** data output from now rx-descriptor to data buffer *****/
/* return value
 *  0: receive is OK
 *  1: nothing has been received
 *  2: receive data is ERROR
 */
static int candy_recv(struct eth_device* dev)
{
	struct candy_hw* hw = (struct candy_hw*)dev->priv;
	volatile DESC *rx_desc =
                (volatile DESC*)(KSEG1ADDR(&rx_descriptor[0]));
	volatile u32 *rcv_cfgr = (volatile u32*)(hw->hw_addr+RCV_CFGR);
	int idx;
	int length;
	u16 status;

	/* if rx DMA is not enable, enable rx DMA & rx DMA burst size setting */
	if( !(*rcv_cfgr & RXE)){
		*rcv_cfgr = RXE | DRBS8;
	}

	for(;;){
		idx = now_rx;
   /* available rx desctiptor does not exist */
		if(idx == next_rx) return 1;

   /* nothing has been received */
		if(!(rx_desc[idx].flag & OWN)) return 1;

   /* receive OK */
		status = rx_desc[idx].flag;
		length = rx_desc[idx].size;
		if( !(status & RXOK)) {
			printf("[Rx Error] status:%04x \n", status);
			return 2;
		}
		/* Pass the packet up to the protocol layyers */
		else {
//			printf("[%d] status:%04x \n", idx, status);
			NetReceive((void*) KSEG1ADDR(NetRxPackets[idx]), length - 4);
		}

   /* free now descriptor */
		rx_desc[idx].flag = 0;
		rx_desc[idx].size = 0;
		rx_desc[idx].addr  = 0;
		now_rx = NEXTRXDESC(idx);
		if( rx_desc_in(hw)) return 2;
		return 0;
	}
}

/**************************************************************************/
/* INIT - Look for an adapter, this routine's visible to the outside      */
/**************************************************************************/
static int candy_init(struct eth_device* dev, bd_t * bd)
{
	struct candy_hw* hw = (struct candy_hw*)dev->priv;
	volatile u32 *macc1 = (volatile u32*)(hw->hw_addr+MACC1);
	int i;

	ether_adr_init(dev);

	ether_mii_reset(hw);
	udelay(100);

	if ( hw->media == MEDIA_NONE )
	  ether_auto_negotiation(hw);

	ether_mac_init(hw);

	ethdsc_init(hw);

	/* make rx descriptor */
	for(i=0;i<CFG_RX_ETH_BUFFER;i++){
		rx_desc_in(hw);
	}

    /* MAC rx enable */
	*macc1 |= SRXEN;

	return 1;
}

/*====================  candy MAC Unit initialize  ===================*/
static int ether_mac_init(struct candy_hw* hw)
{
	volatile u32 *macc1 = (volatile u32*)(hw->hw_addr+MACC1);
	volatile u32 *macc2 = (volatile u32*)(hw->hw_addr+MACC2);
	volatile u32 *ht1 = (volatile u32*)(hw->hw_addr+HT1);
	volatile u32 *ht2 = (volatile u32*)(hw->hw_addr+HT2);
	volatile u32 *afr = (volatile u32*)(hw->hw_addr+AFR);
	volatile u32 *rcv_ctlr = (volatile u32*)(hw->hw_addr+RCV_CTLR);
	volatile u32 *xmt_ctlr = (volatile u32*)(hw->hw_addr+XMT_CTLR);

	if((hw->media ==  MEDIA_100_FULL)||(hw->media ==  MEDIA_10_FULL)){
  // No loopback, No flow control, Full duplex
		*macc1 = PUREP | CRCEN | PADEN | FULLD;
	} else {
  // No loopback, No flow control, Half duplex
		*macc1 = PUREP | CRCEN | PADEN;
	}

	*ht1 = 0x00000000;    // no hash table use
	*ht2 = 0x00000000;
	*afr = PRO;           // All packet recieve mode( Promiscuous mode )
	*rcv_ctlr = UWM | LWM | RX_DRTH8W; // UWM:0x30 LWM:0x10 RX_DRTH:32byte
	*xmt_ctlr = TPTV_aa55 | TX_DRTH8W |  TX_FLTH128B ; // TX_DRTH:32byte,TX_FLTH:128byte
	y_resetw((MCRST|RFRST|TFRST), 10, 0x00000000, macc2);  /* MACC2 soft reset */
	return 0;
}

/*====================  set MAC address  =============================*/
static int ether_adr_init(struct eth_device* dev){
	struct candy_hw* hw = (struct candy_hw*)dev->priv;
	volatile u32 *lsa1 = (volatile u32*)(hw->hw_addr+LSA1);
	volatile u32 *lsa2 = (volatile u32*)(hw->hw_addr+LSA2);
	u32 mac_addr_hi, mac_addr_lo;

        mac_addr_hi  =	(dev->enetaddr[0] <<  8) | (dev->enetaddr[1]      ) ;
        mac_addr_lo  =	(dev->enetaddr[2] << 24) | (dev->enetaddr[3] << 16) |
			(dev->enetaddr[4] <<  8) | (dev->enetaddr[5]      ) ;
	/*
	printf( "ether_adr_init: %x:%x:%x:%x:%x:%x\n",
		mac_addr_hi >> 8,
		mac_addr_hi & 0xFF,
		(mac_addr_lo >> 24) & 0xFF,
		(mac_addr_lo >> 16) & 0xFF,
		(mac_addr_lo >> 8) & 0xFF,
		(mac_addr_lo ) & 0xFF );
	*/
	*lsa2 = mac_addr_hi;
	*lsa1 = mac_addr_lo;
		
	return 0;
}

/*================  MII unit initialization  =======================*/
static void ether_mii_reset(struct candy_hw* hw){
	volatile u32 *macc2 = (volatile u32*)(hw->hw_addr+MACC2);
	volatile u32 *ccr = (volatile u32*)(hw->hw_addr+CCR);
	volatile u32 *miic = (volatile u32*)(hw->hw_addr+MIIC);

	y_resetw((MCRST|RFRST|TFRST), 10, 0x00000000, macc2);  /* MACC2 soft reset */
	y_resetw(SRT, 10, 0x00000000, ccr);  /* CCR soft reset */
	y_resetw((MISRT|CLKS33), 1000, CLKS66, miic);  /* MII block soft reset */
}

/**************************************************************************/
/* MDIO treatment                                                         */
/**************************************************************************/
/***** auto negotiation *****/
static void ether_auto_negotiation(struct candy_hw* hw)
{
	int reg,bAutoNego,i;

	mdio_write(hw, MII_CONTROL, MIICNTL_AUTO);  /* ANEGEN */
	mdio_write(hw, MII_CONTROL, MIICNTL_AUTO | MIICNTL_RST_AUTO);  /* restart auto nego */
	bAutoNego = 0;

	printf( "Auto Negotiation " );
	for(i=0; i < 5000; i++) {
		reg = mdio_read(hw , MII_STATUS);
		if(reg & MIISTAT_AUTO_DONE) {
			bAutoNego = 1;
			break;
		}
		udelay(1000);
	}

	if(!bAutoNego){                         /* Setup 10BASE-T Half duplex */
		mdio_write(hw, MII_CONTROL, 0x0000);
		printf("failed, 10Base-T Half duplex\n");
	}
	else {
		reg = mdio_read(hw, MII_ANLPAR);
		ether_link_status(hw, reg);
	}
}

static void ether_link_status(struct candy_hw* hw, int reg)
{
	if(reg & MII_NWAY_TX_FDX){
		printf("complete: 100Base Full duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_SPEED | MIICNTL_FDX);
		hw->media = MEDIA_100_FULL;
	}
	else if(reg & MII_NWAY_TX){
		printf("complete: 100Base Half duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_SPEED);
		hw->media = MEDIA_100_HALF;
	}
	else if(reg &  MII_NWAY_T_FDX){
		printf("complete: 10Base Full duplex\n");
		//mdio_write(hw, MII_CONTROL, MIICNTL_FDX);
		hw->media = MEDIA_10_FULL;
	}
	else {
		printf("complete: 10Base Half duplex\n");
		//mdio_write(hw, MII_CONTROL, 0x0000);
		hw->media = MEDIA_10_HALF;
	}
}

static unsigned int mdio_read(struct candy_hw* hw, int phyreg_no){
	volatile u32 *madr = (volatile u32*)(hw->hw_addr+MADR);
	volatile u32 *mcmd = (volatile u32*)(hw->hw_addr+MCMD);
	volatile u32 *mind = (volatile u32*)(hw->hw_addr+MIND);
	volatile u32 *mrdd = (volatile u32*)(hw->hw_addr+MRDD);

	/* phy addr & cmd set */
	*madr = (hw->phy_addr <<  FIAD_SHIFT) | phyreg_no;
	*mcmd = RSTAT;

	/* wait not busy = read is comlited */
	while( *mind & BUSY);

	return *mrdd;
}

static void mdio_write(struct candy_hw* hw, int phyreg_no, int data){
	volatile u32 *madr = (volatile u32*)(hw->hw_addr+MADR);
	volatile u32 *mind = (volatile u32*)(hw->hw_addr+MIND);
	volatile u32 *mwtd = (volatile u32*)(hw->hw_addr+MWTD);

	/* phy addr & cmd set */
	*madr = (hw->phy_addr <<  FIAD_SHIFT) | phyreg_no;

	/* data write */
	*mwtd = data;

	/* wait not busy = write is comlited */
	while( *mind & BUSY);
}

/**************************************************************************/
/* Descriptor treatment                                                   */
/**************************************************************************/
static void ethdsc_init(struct candy_hw* hw)
{
	volatile u32 *rcv_dpr = (volatile u32*)(hw->hw_addr+RCV_DPR);
	volatile u32 *xmt_dpr = (volatile u32*)(hw->hw_addr+XMT_DPR);
	volatile DESC *rx_desc =
                (volatile DESC*)(KSEG1ADDR(&rx_descriptor[0]));
	volatile DESC *tx_desc =
				(volatile DESC*)(KSEG1ADDR(&tx_descriptor[0]));
	int i;

//	printf("tx_desc:%x rx_desc:%x \n", &tx_desc[0], &rx_desc[0]);

	for(i=0; i<NUM_RX_DESC; i++){
		rx_desc[i].size = 0;
		rx_desc[i].flag = 0;
		rx_desc[i].addr = 0;
	}
	rx_desc[i].size = 0;
	rx_desc[i].flag = 0;
	rx_desc[i].addr = (virt_to_phys(&rx_desc[0]));

	for(i=0; i<NUM_TX_DESC; i++){
		tx_desc[i].size = 0;
		tx_desc[i].flag = 0;
		tx_desc[i].addr = 0;
	}
	tx_desc[i].size = 0;
	tx_desc[i].flag = 0;
	tx_desc[i].addr = (virt_to_phys(&tx_desc[0])); 

	next_rx = now_rx = 0;

	*rcv_dpr = (virt_to_phys(&rx_desc[0]));
	*xmt_dpr = (virt_to_phys(&tx_desc[0]));
}


/*******************  make next rx-descriptor  ***********************/
/* return value
 *  0: success make next rx descriptor
 *  1: rx descriptor ring is full
 */
static int rx_desc_in(struct candy_hw* hw)
{
	volatile DESC *rx_desc =
                (volatile DESC*)(KSEG1ADDR(&rx_descriptor[0]));
	volatile u32 *rcv_dpr = (volatile u32*)(hw->hw_addr+RCV_DPR);
	volatile u32 *rcv_pdr = (volatile u32*)(hw->hw_addr+RCV_PDR);
	int ri,next;

	/* make next rx-pointer set */
	ri = next_rx;
	next = NEXTRXDESC(ri);
	if(next == now_rx) {
		printf("rx descriptor is full \n");
		return 1;
	}

	rx_desc[ri].size = PKTSIZE_ALIGN;
	rx_desc[ri].flag = DB_LP;
	rx_desc[ri].addr = (virt_to_phys(NetRxPackets[ri]));

	if ( NetRxPackets[ri] < &PktBuf[0] ||
	     NetRxPackets[ri] >= &PktBuf[sizeof(PktBuf)-PKTSIZE_ALIGN] )
	  printf( "rx_desc_in(%d): out of buffer: size: %d, adr=%x (%x< %x <%x)\n",
		  ri, rx_desc[ri].size, NetRxPackets[ri], rx_desc[ri].addr,
		  &PktBuf[0], &PktBuf[sizeof(PktBuf)/PKTSIZE_ALIGN] );

	/* remaing descriptor number(RNOD) is plused 1 */
	*rcv_pdr = 0x1;

	/* if now_rx = ri, Rx-DMA stop. so restart Rx-DMA */
	if(ri == now_rx){
		printf("restart Rx-DMA %d\n", ri );
		*rcv_dpr = (virt_to_phys(&rx_desc[ri]));
	}
	next_rx = next;
	return 0;
}


/**************************************************************************/
/* HALT - Turn off ethernet interface                                     */
/**************************************************************************/
static void candy_halt(struct eth_device* dev){
	struct candy_hw* hw = (struct candy_hw*)dev->priv;
	volatile u32 *macc1 = (volatile u32*)(hw->hw_addr+MACC1);
    /* MAC rx disable */
	*macc1 &= ~SRXEN;
}

static int register_candy( void *hw_addr, uint32_t phy_addr, unsigned char*mac, const char *post )
{
	struct eth_device* dev;
	struct candy_hw* hw;
	int i;

	dev = (struct eth_device*) malloc(sizeof(*dev));
	memset(dev, 0, sizeof(*dev));
	hw = (struct candy_hw*) malloc(sizeof(*hw));
	memset(hw, 0, sizeof(*hw));

	sprintf(dev->name, "NEC-Candy%s", post );
	dev->iobase = 0;
	dev->priv   = hw;
	dev->init   = candy_init;
	dev->halt   = candy_halt;
	dev->send   = candy_send;
	dev->recv   = candy_recv;
	for (i = 0; i < 6; ++i) {
		dev->enetaddr[i] = mac[i];
	}

	hw->hw_addr = hw_addr;
	hw->phy_addr = phy_addr;
	eth_register(dev);

	return 1;
}

int candy_enet_initialize(bd_t *bis){
	return register_candy( (void*)ETH1_BASE, ET0_PHY_ADDR, ((unsigned char*)0xBFFFF108), "" );
}

#endif /* CONFIG_CANDY */
